unit G2App;

interface

uses
  w3system,
  w3application,
  w3components,
  w3graphics,
  w3c.webgl,
  G2Web,
  G2Types,
  G2Utils,
  G2Math,
  G2MeshData,
  G2MeshG2M,
  Shaders;

type
  TGameState = (gsLoading, gsGame);

  TSound = class (TObject)
  private
    _Handle: THandle;
  public
    constructor Create(const FileName: String);
    destructor Destroy; override;
    procedure Play;
    procedure Loop;
    procedure Pause;
    procedure Stop;
  end;

  TParFire = class (TG2S3DParticleFlat)
  private
    _TimeStart: Integer;
    _TimeEnd: Integer;
    _Vel: TG2Vec3;
  public
    procedure Update; override;
    constructor Create; override;
    destructor Destroy; override;
  end;

  TQuickMesh = class
  public
    VB: TG2VertexBuffer;
    IB: TG2IndexBuffer;
    constructor Create;
    destructor Destroy; override;
  end;

  TNodeObjectType = (noUnspecified, noGeometryStatic, noWater);

  TNodeUserData = class
  public
    ObjectType: TNodeObjectType;
  end;

  TMyScene = class (TG2Scene3D)
  public
    ShaderTB: array[0..4, 0..3] of TG2ShaderProgram;
    ShaderTBL: array[0..4, 0..3] of TG2ShaderProgram;
    ShaderTCB: array[0..4, 0..3] of TG2ShaderProgram;
    ShaderTCBL: array[0..4, 0..3] of TG2ShaderProgram;
    ShaderDepthB: array[0..4] of TG2ShaderProgram;
    TranspPassInst: TG2QuickList;
    TranspPassGeom: TG2QuickList;
    constructor Create; override;
    destructor Destroy; override;
    procedure RenderDepth;
    procedure RenderReflection(const WaterInst: TG2S3DMeshInst; const WaterGeom: TG2S3DMeshGeom);
    procedure Render; override;
  end;

  TIntro = class
  private
    procedure PrintLine(const Line: String; const Alpha: TG2Float);
  public
    Lines: array of String;
    LineTime: Integer;
    FadeTime: Integer;
    CurLine: Integer;
    CurTime: Integer;
    constructor Create;
    procedure AddLine(const Line: String);
    procedure Render;
    procedure Update;
    procedure Reset;
  end;

  TMyApp = class (TObject)
  public
    State: TGameState;
    Font: TG2Font;
    Scene: TMyScene;
    Sky: TQuickMesh;
    Landscape: TG2S3DMesh;
    LandscapeInst: TG2S3DMeshInst;
    ShaderWater: TG2ShaderProgram;
    ShaderSky: TG2ShaderProgram;
    TexFireball: TG2Texture2D;
    TexSky: TG2TextureCube;
    TexNM: TG2Texture2D;
    TexWaterRef: TG2Texture2DRT;
    TexDepth: TG2Texture2DRT;
    Music: TSound;
    Path: array of TG2Vec3;
    CurPos: TG2Float;
    LookPos: TG2Float;
    Emmitters: array of TG2Vec3;
    EmmitTime: Integer;
    FadeTime: Integer;
    Intro: TIntro;
    procedure DrawPrimitive(const PrimitiveCount: Integer);
    procedure DrawPrimitiveIndexed(const PrimitiveCount: Integer);
    procedure Initialize;
    procedure Finalize;
    procedure Render;
    procedure Update;
    procedure StartDemo;
    procedure RenderSky(const SkyW, SkyV, SkyP: TG2Mat);
    procedure RenderLandscape;
    procedure MakeSky;
    procedure MakePath;
    procedure MakeEmmitters;
    procedure UpdateEmmitters;
    function AdvancePos(const StartPos, Amount: Single): Single;
    function GetPathPos(const Pos: Single): TG2Vec3;
  end;

var App: TMyApp;

implementation

//TSound BEGIN
constructor TSound.Create(const FileName: String);
begin
  inherited Create;
  asm
    @_Handle = new Audio();
  end;
  _Handle.src := FileName;
end;

destructor TSound.Destroy;
begin
  Stop;
  _Handle := nil;
  inherited Destroy;
end;

procedure TSound.Play;
begin
  _Handle.loop := False;
  _Handle.play();
end;

procedure TSound.Loop;
begin
  _Handle.loop := True;
  _Handle.play();
end;

procedure TSound.Pause;
begin
  _Handle.pause();
end;

procedure TSound.Stop;
begin
  _Handle.pause();
  _Handle.currentTime := 0;
end;
//TSound END

//TParFire BEGIN
procedure TParFire.Update;
  var t, s: Float;
  var c: Float;
begin
  t := (G2Time - _TimeStart) / _TimeEnd;
  if t < 1 then
  begin
    Pos := Pos + _Vel;
    Pos.y := Pos.y + 0.03;
    c := Sin(t * Pi);
    Color := G2Vec4(c, c, c, 1);
    s := G2LerpFloat(7, 2, t);
    MakeBillboard(App.Scene.V, s, s, G2PiTime);
  end
  else
  Die;
end;

constructor TParFire.Create;
begin
  inherited Create;
  Texture := App.TexFireball;
  _TimeStart := G2Time;
  _TimeEnd := 3000;
  _Vel := G2RandomSpherePoint * G2Random(101) * 0.0002;
  BlendMode := bmAdd;
  DepthSorted := False;
end;

destructor TParFire.Destroy;
begin
  inherited Destroy;
end;
//TParFire END

//TQuickMesh BEGIN
constructor TQuickMesh.Create;
begin
  inherited Create;
  VB := TG2VertexBuffer.Create;
  IB := TG2IndexBuffer.Create;
end;

destructor TQuickMesh.Destroy;
begin
  IB.Free;
  VB.Free;
  inherited Destroy;
end;
//TQuickMesh END

//TMyScene BEGIN
constructor TMyScene.Create;
  var i, j: Integer;
begin
  inherited Create;
  for i := 0 to 4 do
  begin
    for j := 0 to 3 do
    begin
      ShaderTB[i, j] := TG2ShaderProgram.Create; ShaderTB[i, j].Parent := Self; ShaderTB[i, j].Compile(VS_Scene(i, j, False, False), PS_Scene(j, False, False));
      ShaderTBL[i, j] := TG2ShaderProgram.Create; ShaderTBL[i, j].Parent := Self; ShaderTBL[i, j].Compile(VS_Scene(i, j, False, True), PS_Scene(j, False, True));
      ShaderTCB[i, j] := TG2ShaderProgram.Create; ShaderTCB[i, j].Parent := Self; ShaderTCB[i, j].Compile(VS_Scene(i, j, True, False), PS_Scene(j, True, False));
      ShaderTCBL[i, j] := TG2ShaderProgram.Create; ShaderTCBL[i, j].Parent := Self; ShaderTCBL[i, j].Compile(VS_Scene(i, j, True, True), PS_Scene(j, True, True));
    end;
    ShaderDepthB[i] := TG2ShaderProgram.Create; ShaderDepthB[i].Parent := Self; ShaderDepthB[i].Compile(VS_Depth(i), PS_Depth);
  end;
end;

destructor TMyScene.Destroy;
  var i, j: Integer;
begin
  for i := 0 to 4 do
  begin
    for j := 0 to 3 do
    begin
      ShaderTB[i, j].Free;
      ShaderTBL[i, j].Free;
      ShaderTCB[i, j].Free;
      ShaderTCBL[i, j].Free;
    end;
    ShaderDepthB[i].Free;
  end;
end;

procedure TMyScene.RenderDepth;
  var i, j, g, m: Integer;
  var Mesh: TG2S3DMesh;
  var Inst: TG2S3DMeshInst;
  var Geom: TG2S3DMeshGeom;
  var W, VP, WVP, InvV: TG2Mat;
  var VB: TG2VertexBuffer;
  var ShaderMethod: Integer;
  var Shader: TG2ShaderProgram;
  var v3: TG2Vec3;
begin
  g2.gfx.RenderTarget := App.TexDepth;
  gl.clearColor(1, 1, 1, 1);
  gl.clear(gl.COLOR_BUFFER_BIT or gl.DEPTH_BUFFER_BIT);
  gl.Enable(gl.CULL_FACE);
  gl.cullFace(gl.BACK);
  VP := V * G2MatScaling(1, -1, 1) * P;
  v3 := G2Vec3(-V.e30, -V.e31, -V.e32);
  v3 := G2Vec3MatMul3x3(v3, G2MatTranspose(V));
  for i := 0 to MeshInstCount - 1 do
  begin
    Inst := MeshInst[i];
    Mesh := Inst.Mesh;
    if Frustum.CheckBox(G2BoxToAABox(Inst.BBox)) <> fcOutside then
    for g := 0 to Mesh.GeomCount - 1 do
    begin
      Geom := Mesh.Geoms[g];
      if TNodeUserData(Inst.UserData[Geom.NodeID]).ObjectType = noGeometryStatic then
      begin
        if Geom.Skinned then
        begin
          VB := TG2S3DGeomDataSkinned(Geom.Data).VB;
          W := Inst.Transform;
          ShaderMethod := TG2S3DGeomDataSkinned(Geom.Data).MaxWeights;
        end
        else
        begin
          VB := TG2S3DGeomDataStatic(Geom.Data).VB;
          W := Inst.Transforms[Geom.NodeID].TransformCom * Inst.Transform;
          ShaderMethod := 0;
        end;
        WVP := W * VP;
        Geom.IB.Bind;
        Shader := ShaderDepthB[ShaderMethod];
        Shader.Use;
        if Geom.Skinned then
        begin
          for j := 0 to TG2S3DGeomDataSkinned(Geom.Data).BoneCount - 1 do
          gl.uniformMatrix4fv(
            Shader.GetShaderUniform('SkinPallete[' + IntToStr(j) + ']'),
            False, G2MatTranspose(Inst.Skins[g].Transforms[j]).ToFloat32Array
          );
        end;
        gl.uniform2f(Shader.GetShaderUniform('DepthClamp'), 1, 1000);
        gl.uniform3fv(Shader.GetShaderUniform('CamPos'), v3.ToFloat32Array);
        gl.uniformMatrix4fv(Shader.GetShaderUniform('W'), False, G2MatTranspose(W).ToFloat32Array);
        gl.uniformMatrix4fv(Shader.GetShaderUniform('WVP'), False, G2MatTranspose(WVP).ToFloat32Array);
        VB.Bind;
        gl.DrawElements(
          gl.TRIANGLES,
          Geom.FCount * 3,
          gl.UNSIGNED_SHORT,
          0
        );
        VB.Unbind;
        Geom.IB.Unbind;
      end;
    end;
  end;
  g2.gfx.RenderTarget := nil;
end;

procedure TMyScene.RenderReflection(const WaterInst: TG2S3DMeshInst; const WaterGeom: TG2S3DMeshGeom);
  var i, j, g, m, CurStage: Integer;
  var Mesh: TG2S3DMesh;
  var Inst: TG2S3DMeshInst;
  var Geom: TG2S3DMeshGeom;
  var W, VP, WVP, InvW: TG2Mat;
  var VB: TG2VertexBuffer;
  var ShaderMethod: Integer;
  var Tex: array[0..1] of TG2Texture2D;
  var Shader: TG2ShaderProgram;
  var Material: TG2S3DMeshMaterial;
  var ClipPlane: TG2Vec4;
  var WaterLevel: TG2Float;
  procedure SetDiffuseTexture(const Texture: TG2Texture2D; const ID: Integer);
  begin
    if Texture <> nil then
    begin
      gl.uniform1i(Shader.GetShaderUniform('Tex' + IntToStr(ID)), CurStage);
      if Tex[CurStage] <> Texture then
      begin
        Tex[CurStage] := Texture;
        Tex[CurStage].Bind(CurStage);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      end;
      Inc(CurStage);
    end;
  end;
begin
  WaterLevel := WaterInst.Transforms[WaterGeom.NodeID].TransformCom.GetTranslation.y;
  ClipPlane := G2Vec4(0, -1, 0, WaterLevel);
  g2.Gfx.RenderTarget := App.TexWaterRef;
  gl.clearColor(1, 1, 1, 0);
  gl.clear(gl.COLOR_BUFFER_BIT or gl.DEPTH_BUFFER_BIT);
  App.RenderSky(G2MatScaling(600, -600, 600), V * G2MatScaling(1, -1, 1), P);
  gl.Enable(gl.CULL_FACE);
  VP := V * G2MatScaling(1, -1, 1) * P;
  InvW := G2MatScaling(1, -1, 1) * G2MatTranslation(0, WaterLevel * 2 + 0.4, 0);
  for i := 0 to MeshInstCount - 1 do
  begin
    Inst := MeshInst[i];
    Mesh := Inst.Mesh;
    if Frustum.CheckBox(G2BoxToAABox(Inst.BBox.Transform(InvW))) <> fcOutside then
    for g := 0 to Mesh.GeomCount - 1 do
    begin
      Geom := Mesh.Geoms[g];
      if TNodeUserData(Inst.UserData[Geom.NodeID]).ObjectType = noGeometryStatic then
      begin
        if Geom.Skinned then
        begin
          VB := TG2S3DGeomDataSkinned(Geom.Data).VB;
          W := Inst.Transform;
          ShaderMethod := TG2S3DGeomDataSkinned(Geom.Data).MaxWeights;
        end
        else
        begin
          VB := TG2S3DGeomDataStatic(Geom.Data).VB;
          W := Inst.Transforms[Geom.NodeID].TransformCom * Inst.Transform;
          ShaderMethod := 0;
        end;
        W := W * InvW;
        WVP := G2MatTranspose(W * VP);
        Geom.IB.Bind;
        for m := 0 to Geom.GCount - 1 do
        begin
          Material := Inst.Materials[Geom.Groups[m].Material];
          if Material.ChannelCount > 0 then
          begin
            CurStage := 0;
            if Material.Channels[0].MapLight <> nil then
            begin
              Shader := ShaderTCBL[ShaderMethod, Material.ChannelCount - 1];
              Shader.Use;
              gl.uniform1i(Shader.GetShaderUniform('TexLight'), CurStage);
              if Tex[CurStage] <> Material.Channels[0].MapLight then
              begin
                Tex[CurStage] := Material.Channels[0].MapLight;
                Tex[CurStage].Bind(CurStage);
                gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
                gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
              end;
              Inc(CurStage);
            end
            else
            begin
              Shader := ShaderTCB[ShaderMethod, Material.ChannelCount - 1];
              Shader.Use;
            end;
            for i := 0 to Material.ChannelCount - 1 do
            SetDiffuseTexture(Material.Channels[i].MapDiffuse, i);
          end
          else
          begin
            Shader := ShaderTCB[0, 0];
            Shader.Use;
            gl.uniform1i(Shader.GetShaderUniform('Tex0'), 0);
          end;
          if Geom.Skinned then
          begin
            for j := 0 to TG2S3DGeomDataSkinned(Geom.Data).BoneCount - 1 do
            gl.uniformMatrix4fv(
              Shader.GetShaderUniform('SkinPallete[' + IntToStr(j) + ']'),
              False, G2MatTranspose(Inst.Skins[g].Transforms[j]).ToFloat32Array
            );
          end;
          gl.uniform4fv(Shader.GetShaderUniform('ClipPlane'), ClipPlane.ToFloat32Array);
          gl.uniformMatrix4fv(Shader.GetShaderUniform('WVP'), False, WVP.ToFloat32Array);
          gl.uniformMatrix4fv(Shader.GetShaderUniform('W'), False, W.ToFloat32Array);
          gl.uniform4fv(Shader.GetShaderUniform('LightAmbient'), Ambient.ToFloat32Array);
          VB.Bind;
          gl.DrawElements(
            gl.TRIANGLES,
            Geom.Groups[m].FaceCount * 3,
            gl.UNSIGNED_SHORT,
            Geom.Groups[m].FaceStart * 6
          );
          VB.Unbind;
        end;
        Geom.IB.Unbind;
      end;
    end;
  end;
  g2.Gfx.RenderTarget := nil;
end;

procedure TMyScene.Render;
  var i, j, g, m: Integer;
  var f: Single;
  var CurStage: Integer;
  var W, VP, WVP, TexCoordTransform, TexProj: TG2Mat;
  var v3: TG2Vec3;
  var Mesh: TG2S3DMesh;
  var Inst: TG2S3DMeshInst;
  var Geom: TG2S3DMeshGeom;
  var Material: TG2S3DMeshMaterial;
  var PrevDepthEnable: Boolean;
  var VB: TG2VertexBuffer;
  var ShaderMethod: Integer;
  var Tex: array[0..4] of TG2Texture2D;
  var Shader: TG2ShaderProgram;
  procedure SetDiffuseTexture(const Texture: TG2Texture2D; const ID: Integer);
  begin
    if Texture <> nil then
    begin
      gl.uniform1i(Shader.GetShaderUniform('Tex' + IntToStr(ID)), CurStage);
      if Tex[CurStage] <> Texture then
      begin
        Tex[CurStage] := Texture;
        Tex[CurStage].Bind(CurStage);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      end;
      Inc(CurStage);
    end;
  end;
begin
  g2.RenderControl := nil;
  TranspPassInst.Clear;
  TranspPassGeom.Clear;
  g2.Gfx.BlendMode := bmNormal;
  Frustum.Update;
  PrevDepthEnable := g2.Gfx.DepthEnable;
  g2.Gfx.DepthEnable := True;
  gl.Enable(gl.CULL_FACE);
  VP := V * P;
  for i := 0 to High(Tex) do
  Tex[i] := nil;
  for i := 0 to MeshInstCount - 1 do
  begin
    Inst := MeshInst[i];
    Mesh := Inst.Mesh;
    if Frustum.CheckBox(G2BoxToAABox(Inst.BBox)) <> fcOutside then
    for g := 0 to Mesh.GeomCount - 1 do
    begin
      Geom := Mesh.Geoms[g];
      if TNodeUserData(Inst.UserData[Geom.NodeID]).ObjectType = noWater then
      begin
        TranspPassInst.Add(Inst);
        TranspPassGeom.Add(Geom);
      end
      else
      begin
        g2.Gfx.BlendMode := bmDisable;
        if Geom.Skinned then
        begin
          VB := TG2S3DGeomDataSkinned(Geom.Data).VB;
          W := Inst.Transform;
          ShaderMethod := TG2S3DGeomDataSkinned(Geom.Data).MaxWeights;
        end
        else
        begin
          VB := TG2S3DGeomDataStatic(Geom.Data).VB;
          W := Inst.Transforms[Geom.NodeID].TransformCom * Inst.Transform;
          ShaderMethod := 0;
        end;
        WVP := G2MatTranspose(W * VP);
        Geom.IB.Bind;
        for m := 0 to Geom.GCount - 1 do
        begin
          Material := Inst.Materials[Geom.Groups[m].Material];
          if Material.ChannelCount > 0 then
          begin
            CurStage := 0;
            if Material.Channels[0].MapLight <> nil then
            begin
              Shader := ShaderTBL[ShaderMethod, Material.ChannelCount - 1];
              Shader.Use;
              gl.uniform1i(Shader.GetShaderUniform('TexLight'), CurStage);
              if Tex[CurStage] <> Material.Channels[0].MapLight then
              begin
                Tex[CurStage] := Material.Channels[0].MapLight;
                Tex[CurStage].Bind(CurStage);
                gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
                gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
              end;
              Inc(CurStage);
            end
            else
            begin
              Shader := ShaderTB[ShaderMethod, Material.ChannelCount - 1];
              Shader.Use;
            end;
            for i := 0 to Material.ChannelCount - 1 do
            SetDiffuseTexture(Material.Channels[i].MapDiffuse, i);
          end
          else
          begin
            Shader := ShaderTB[0, 0];
            Shader.Use;
            gl.uniform1i(Shader.GetShaderUniform('Tex0'), 0);
          end;
          if Geom.Skinned then
          begin
            for j := 0 to TG2S3DGeomDataSkinned(Geom.Data).BoneCount - 1 do
            gl.uniformMatrix4fv(
              Shader.GetShaderUniform('SkinPallete[' + IntToStr(j) + ']'),
              False, G2MatTranspose(Inst.Skins[g].Transforms[j]).ToFloat32Array
            );
          end;
          gl.uniformMatrix4fv(Shader.GetShaderUniform('WVP'), False, WVP.ToFloat32Array);
          gl.uniform4fv(Shader.GetShaderUniform('LightAmbient'), Ambient.ToFloat32Array);
          VB.Bind;
          gl.DrawElements(
            gl.TRIANGLES,
            Geom.Groups[m].FaceCount * 3,
            gl.UNSIGNED_SHORT,
            Geom.Groups[m].FaceStart * 6
          );
          VB.Unbind;
        end;
        Geom.IB.Unbind;
      end;
    end;
  end;
  for i := 0 to TranspPassInst.Count - 1 do
  begin
    Inst := TG2S3DMeshInst(TranspPassInst[i]);
    Geom := TG2S3DMeshGeom(TranspPassGeom[i]);
    if TNodeUserData(Inst.UserData[Geom.NodeID]).ObjectType = noWater then
    begin
      RenderReflection(Inst, Geom);
      g2.Gfx.BlendMode := bmNormal;
      VB := TG2S3DGeomDataStatic(Geom.Data).VB;
      W := Inst.Transforms[Geom.NodeID].TransformCom * Inst.Transform;
      WVP := W * VP;
      App.ShaderWater.Use;
      TexCoordTransform := G2MatRotationZ(Pi * 0.25);
      TexProj := WVP * App.TexWaterRef.Projection;
      v3 := G2Vec3(-V.e30, -V.e31, -V.e32);
      v3 := G2Vec3MatMul3x3(v3, G2MatTranspose(V));
      gl.uniform2f(App.ShaderWater.GetShaderUniform('DepthClamp'), 1, 1000);
      gl.uniform3f(App.ShaderWater.GetShaderUniform('CamPos'), v3.x, v3.y, v3.z);
      gl.uniform4f(App.ShaderWater.GetShaderUniform('WaterOffset'), G2TimeInterval(18001), G2TimeInterval(17123), -G2TimeInterval(19002), G2TimeInterval(17527));
      gl.uniformMatrix4fv(App.ShaderWater.GetShaderUniform('W'), False, G2MatTranspose(W).ToFloat32Array);
      gl.uniformMatrix4fv(App.ShaderWater.GetShaderUniform('WVP'), False, G2MatTranspose(WVP).ToFloat32Array);
      gl.uniformMatrix4fv(App.ShaderWater.GetShaderUniform('TexCoordTransform'), False, G2MatTranspose(TexCoordTransform).ToFloat32Array);
      gl.uniformMatrix4fv(App.ShaderWater.GetShaderUniform('TexProj'), False, G2MatTranspose(TexProj).ToFloat32Array);
      App.TexSky.Bind(0);
      gl.uniform1i(App.ShaderWater.GetShaderUniform('Tex0'), 0);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_LINEAR);
      App.TexNM.Bind(1);
      gl.uniform1i(App.ShaderWater.GetShaderUniform('Tex1'), 1);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_LINEAR);
      App.TexWaterRef.Bind(2);
      gl.uniform1i(App.ShaderWater.GetShaderUniform('Tex2'), 2);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
      App.TexDepth.Bind(3);
      gl.uniform1i(App.ShaderWater.GetShaderUniform('Tex3'), 3);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
      gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
      if Geom.GCount > 0 then
      begin
        Material := Inst.Materials[Geom.Groups[0].Material];
        if Material.ChannelCount > 0 then
        begin
          if Material.Channels[0].MapLight <> nil then
          begin
            Material.Channels[0].MapLight.Bind(4);
            gl.uniform1i(App.ShaderWater.GetShaderUniform('TexLight'), 4);
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
          end;
        end;
      end;
      VB.Bind;
      Geom.IB.Bind;
      gl.DrawElements(
        gl.TRIANGLES,
        Geom.FCount * 3,
        gl.UNSIGNED_SHORT,
        0
      );
      Geom.IB.Unbind;
      VB.Unbind;
    end;
  end;
  gl.Disable(gl.CULL_FACE);
  RenderParticles;
  g2.Gfx.DepthEnable := PrevDepthEnable;
end;
//TMyScene END

//TIntro BEGIN
procedure TIntro.PrintLine(const Line: String; const Alpha: TG2Float);
  var w, h, s, sx, sy: TG2Float;
begin
  w := App.Font.TextWidth(Line);
  h := App.Font.TextHeight(Line);
  sx := g2.Params.Width / w;
  sy := g2.Params.Height / h;
  if (sx < 1) or (sy < 1) then
  begin
    if sx < sy then s := sx else s := sy;
  end
  else s := 1;
  sx := w * s;
  sy := h * s;
  App.Font.Print(
    (g2.Params.Width - sx) * 0.5,
    (g2.Params.Height - sy) * 0.5,
    s, s, G2Vec4(1, 1, 1, Alpha),
    Line, bmNormal, tfLinear
  );
end;

constructor TIntro.Create;
begin
  inherited Create;
  Lines.SetLength(0);
  LineTime := 2000;
  FadeTime := 2000;
  CurLine := 0;
end;

procedure TIntro.AddLine(const Line: String);
begin
  Lines.SetLength(Length(Lines) + 1);
  Lines[High(Lines)] := Line;
end;

procedure TIntro.Render;
  var t: Integer;
begin
  if CurLine < Length(Lines) then
  begin
    t := G2Time - CurTime;
    if t < FadeTime then
    begin
      PrintLine(Lines[CurLine], t / FadeTime);
    end
    else if t < FadeTime + LineTime then
    begin
      PrintLine(Lines[CurLine], 1);
    end
    else if t < FadeTime * 2 + LineTime then
    begin
      PrintLine(Lines[CurLine], 1 - (t - (FadeTime + LineTime)) / FadeTime);
    end;
  end;
end;

procedure TIntro.Update;
  var t: Integer;
begin
  if CurLine < Length(Lines) then
  begin
    t := G2Time - CurTime;
    if t > FadeTime * 2 + LineTime then
    begin
      Inc(CurLine);
      CurTime := G2Time;
    end;
  end;
end;

procedure TIntro.Reset;
begin
  CurLine := 0;
  CurTime := G2Time;
end;
//TIntro END

//TMyApp BEGIN
procedure TMyApp.DrawPrimitive(const PrimitiveCount: Integer);
begin
  g2.Gfx.gl.drawArrays(g2.Gfx.gl.TRIANGLES, 0, PrimitiveCount * 3);
end;

procedure TMyApp.DrawPrimitiveIndexed(const PrimitiveCount: Integer);
begin
  g2.Gfx.gl.DrawElements(g2.Gfx.gl.TRIANGLES, PrimitiveCount * 3, g2.Gfx.gl.UNSIGNED_SHORT, 0);
end;

procedure TMyApp.Initialize;
begin
  PreventCaching := False;
  State := gsLoading;
  ShaderWater := TG2ShaderProgram.Create;
  ShaderWater.Compile(VS_Water, PS_Water);
  MakeSky;
  Font := TG2Font.Create;
  Font.Load('Font0.g2fh', 'Font0.png');
  TexFireball := TG2Texture2D.Create;
  TexFireball.Load('Fireball.png', tuUsage3D);
  TexSky := TG2TextureCube.Create;
  TexSky.Load(
    'PosX.png',
    'NegX.png',
    'PosY.png',
    'NegY.png',
    'PosZ.png',
    'NegZ.png'
  );
  TexNM := TG2Texture2D.Create;
  TexNM.Load('nm1.jpg');
  TexWaterRef := TG2Texture2DRT.Create;
  TexWaterRef.Make(512, 512);
  TexDepth := TG2Texture2DRT.Create;
  TexDepth.Make(512, 512);
  Scene := TMyScene.Create;
  Scene.Frustum.RefP := Scene.P.CopyRef;
  Scene.Frustum.RefV := Scene.V.CopyRef;
  Scene.Ambient := G2Vec4(0, 0, 0, 0);
  Landscape := TG2S3DMesh.Create(Scene);
  Landscape.Load('scene.g2m');
  Music := TSound.Create('Music1.wav');
  Intro := TIntro.Create;
  Intro.AddLine('Pascode presents');
  Intro.AddLine('Afternoon Walk demoscene by Dan');
  Intro.AddLine('Powered by g2mp game development framework');
  Intro.AddLine('Exclusively for Smart Mobile Studio Graphics Contest!');
  Intro.AddLine('Music by Stephane Picq');
end;

procedure TMyApp.Finalize;
begin
  Scene.Free;
  ShaderWater.Free;
  Music.Free;
end;

procedure TMyApp.Render;
  var i: Integer;
  var Rot, Ang, Radius0, Radius1, s, c, a0, a1, x, y: TG2Float;
  var v0r0, v0r1, v1r0, v1r1: TG2Vec2;
  var Segments: Integer;
begin
  case State of
    gsLoading:
    begin
      g2.ClearColor := G2Vec4(0, 0, 0, 1);
      g2.Gfx.gl.clear(g2.Gfx.gl.COLOR_BUFFER_BIT);
      Segments := 40;
      Radius0 := 32;
      Radius1 := 48;
      Rot := G2PiTime;
      Ang := Rot;
      x := g2.Params.Width * 0.5;
      y := g2.Params.Height * 0.5;
      G2SinCos(Ang, s, c);
      g2.PrimBegin(ptTriangles, bmNormal);
      v1r0 := G2Vec2(c * Radius0 + x, s * Radius0 + y);
      v1r1 := G2Vec2(c * Radius1 + x, s * Radius1 + y);
      a1 := 0;
      for i := 1 to Segments do
      begin
        v0r0 := v1r0; v0r1 := v1r1; a0 := a1;
        Ang := TwoPi * i / Segments + Rot;
        G2SinCos(Ang, s, c);
        v1r0 := G2Vec2(c * Radius0 + x, s * Radius0 + y);
        v1r1 := G2Vec2(c * Radius1 + x, s * Radius1 + y);
        a1 := i / Segments;
        g2.PrimAdd(v0r0, G2Vec4(1, 1, 1, a0));
        g2.PrimAdd(v0r1, G2Vec4(1, 1, 1, a0));
        g2.PrimAdd(v1r0, G2Vec4(1, 1, 1, a1));
        g2.PrimAdd(v1r0, G2Vec4(1, 1, 1, a1));
        g2.PrimAdd(v0r1, G2Vec4(1, 1, 1, a0));
        g2.PrimAdd(v1r1, G2Vec4(1, 1, 1, a1));
      end;
    end;
    gsGame:
    begin
      Scene.P := G2MatProj(HalfPi, g2.Params.Width / g2.Params.Height, 1, 1000);;
      g2.Gfx.DepthEnable := True;
      Scene.RenderDepth;
      Scene.Render;
      RenderSky(G2MatScaling(600, 600, 600), Scene.V, Scene.P);
      g2.Gfx.DepthEnable := False;
      i := G2Time - FadeTime;
      if i < 3000 then
      begin
        s := i / 3000;
        g2.PrimRect(0, 0, g2.Params.Width, g2.Params.Height, G2Vec4(0, 0, 0, 1 - s));
      end;
      Intro.Render;
    end;
  end;
end;

procedure TMyApp.Update;
begin
  case State of
    gsLoading:
    begin
      if (g2.LoadingItems = 0) then
      begin
        StartDemo;
        State := gsGame;
      end;
    end;
    gsGame:
    begin
      Intro.Update;
      UpdateEmmitters;
      CurPos := AdvancePos(CurPos, 0.1);
      LookPos := AdvancePos(CurPos, 0.5);
      if CurPos > Length(Path) then
      CurPos := CurPos - Length(Path);
      if LookPos > Length(Path) then
      LookPos := LookPos - Length(Path);
      Scene.V := G2MatView(GetPathPos(CurPos), GetPathPos(LookPos), G2Vec3(0, 1, 0));
      Scene.Update;
    end;
  end;
end;

procedure TMyApp.StartDemo;
  var i, j, n: Integer;
  var StrArr: TG2StrArr;
begin
  LandscapeInst := Landscape.NewInst;
  MakePath;
  MakeEmmitters;
  EmmitTime := G2Time;
  FadeTime := G2Time;
  CurPos := 0;
  LookPos := CurPos + 0.4;
  for i := 0 to Scene.MeshInstCount - 1 do
  begin
    for j := 0 to Scene.MeshInst[i].Mesh.NodeCount - 1 do
    begin
      Scene.MeshInst[i].UserData[j] := TNodeUserData.Create;
      TNodeUserData(Scene.MeshInst[i].UserData[j]).ObjectType := noUnspecified;
      StrArr := G2StrExplode(Scene.MeshInst[i].Mesh.Nodes[j].Name, '_');
      for n := 0 to Length(StrArr) - 1 do
      if StrArr[n] = '[w]' then
      begin
        TNodeUserData(Scene.MeshInst[i].UserData[j]).ObjectType := noWater;
      end
      else if StrArr[n] = '[r]' then
      begin
        TNodeUserData(Scene.MeshInst[i].UserData[j]).ObjectType := noGeometryStatic;
      end;
    end;
  end;
  Intro.Reset;
  Music.Loop;
end;

procedure TMyApp.RenderSky(const SkyW, SkyV, SkyP: TG2Mat);
  var WVP, V: TG2Mat;
begin
  V := SkyV; V.e30 := 0; V.e31 := 0; V.e32 := 0;
  WVP := SkyW * V * SkyP;
  ShaderSky.Use;
  g2.Gfx.gl.uniformMatrix4fv(ShaderSky.GetShaderUniform('WVP'), False, G2MatTranspose(WVP).ToFloat32Array);
  TexSky.Bind(0);
  g2.Gfx.gl.uniform1i(ShaderSky.GetShaderUniform('Tex0'), 0);
  Sky.VB.Bind;
  Sky.IB.Bind;
  DrawPrimitiveIndexed(12);
  Sky.IB.Unbind;
  Sky.IB.Unbind;
end;

procedure TMyApp.RenderLandscape;
begin

end;

procedure TMyApp.MakeSky;
  var Decl: TG2VBDecl;
begin
  Decl.SetLength(1);
  Decl[0].Element := vbPosition; Decl[0].Count := 3;
  Sky := TQuickMesh.Create;
  Sky.VB.Initialize(Decl, 8);
  Sky.VB.Lock;
  Sky.VB.WriteVec3(G2Vec3(-0.5, -0.5, -0.5));
  Sky.VB.WriteVec3(G2Vec3(-0.5, -0.5, 0.5));
  Sky.VB.WriteVec3(G2Vec3(0.5, -0.5, 0.5));
  Sky.VB.WriteVec3(G2Vec3(0.5, -0.5, -0.5));
  Sky.VB.WriteVec3(G2Vec3(-0.5, 0.5, -0.5));
  Sky.VB.WriteVec3(G2Vec3(-0.5, 0.5, 0.5));
  Sky.VB.WriteVec3(G2Vec3(0.5, 0.5, 0.5));
  Sky.VB.WriteVec3(G2Vec3(0.5, 0.5, -0.5));
  Sky.VB.UnLock;
  Sky.IB.Initialize(36);
  Sky.IB.Lock;
  Sky.IB.WriteIntU16(1);
  Sky.IB.WriteIntU16(3);
  Sky.IB.WriteIntU16(0);
  Sky.IB.WriteIntU16(2);
  Sky.IB.WriteIntU16(3);
  Sky.IB.WriteIntU16(1);
  Sky.IB.WriteIntU16(7);
  Sky.IB.WriteIntU16(4);
  Sky.IB.WriteIntU16(0);
  Sky.IB.WriteIntU16(3);
  Sky.IB.WriteIntU16(7);
  Sky.IB.WriteIntU16(0);
  Sky.IB.WriteIntU16(6);
  Sky.IB.WriteIntU16(5);
  Sky.IB.WriteIntU16(4);
  Sky.IB.WriteIntU16(7);
  Sky.IB.WriteIntU16(6);
  Sky.IB.WriteIntU16(4);
  Sky.IB.WriteIntU16(5);
  Sky.IB.WriteIntU16(6);
  Sky.IB.WriteIntU16(2);
  Sky.IB.WriteIntU16(1);
  Sky.IB.WriteIntU16(5);
  Sky.IB.WriteIntU16(2);
  Sky.IB.WriteIntU16(4);
  Sky.IB.WriteIntU16(5);
  Sky.IB.WriteIntU16(1);
  Sky.IB.WriteIntU16(0);
  Sky.IB.WriteIntU16(4);
  Sky.IB.WriteIntU16(1);
  Sky.IB.WriteIntU16(6);
  Sky.IB.WriteIntU16(7);
  Sky.IB.WriteIntU16(3);
  Sky.IB.WriteIntU16(2);
  Sky.IB.WriteIntU16(6);
  Sky.IB.WriteIntU16(3);
  Sky.IB.UnLock;
  ShaderSky := TG2ShaderProgram.Create;
  ShaderSky.Compile(VS_Sky, PS_Sky);
end;

procedure TMyApp.MakePath;
  function GetNode(const NodeName: String; var NodeFound: Boolean): TG2Mat;
    var i: Integer;
  begin
    for i := 0 to Landscape.NodeCount - 1 do
    if Landscape.Nodes[i].Name = NodeName then
    begin
      Result := Landscape.Nodes[i].Transform;
      NodeFound := True;
      Exit;
    end;
    NodeFound := False;
  end;
  var b: Boolean;
  var w: TG2QuickListVariant;
  var i: Integer;
  var m: TG2Mat;
  var vr: Variant;
begin
  w.Clear;
  i := 0;
  repeat
    m := GetNode('Path_' + IntToStr(i), b);
    if b then
    begin
      w.Add(m);
      Inc(i);
    end;
  until not b;
  Path.SetLength(w.Count);
  for i := 0 to w.Count - 1 do
  begin
    vr := w[i];
    asm @m = @vr; end;
    Path[i].x := m.e30;
    Path[i].y := m.e31;
    Path[i].z := m.e32;
  end;
end;

procedure TMyApp.MakeEmmitters;
  var i, j, n: Integer;
  var ParamArr: TG2StrArr;
  var Transform: TG2Mat;
begin
  for i := 0 to Landscape.NodeCount - 1 do
  begin
    ParamArr := G2StrExplode(Landscape.Nodes[i].Name, '_');
    for j := 0 to High(ParamArr) do
    begin
      if ParamArr[j] = 'Flame' then
      begin
        Emmitters.SetLength(Length(Emmitters) + 1);
        Transform := Landscape.Nodes[i].Transform;
        n := Landscape.Nodes[i].OwnerID;
        while n > -1 do
        begin
          Transform := Transform * Landscape.Nodes[n].Transform;
          n := Landscape.Nodes[n].OwnerID;
        end;
        Emmitters[High(Emmitters)].x := Transform.e30;
        Emmitters[High(Emmitters)].y := Transform.e31;
        Emmitters[High(Emmitters)].z := Transform.e32;
      end;
    end;
  end;
end;

procedure TMyApp.UpdateEmmitters;
  var i: Integer;
  var p: TParFire;
begin
  if (G2Time - EmmitTime > 200) then
  begin
    for i := 0 to High(Emmitters) do
    begin
      p := TParFire.Create;
      p.Pos := Emmitters[i] + G2RandomSpherePoint * (G2Random(101) * 0.01) * 0.3;
      Scene.ParticleAdd(p);
    end;
    EmmitTime := G2Time;
  end;
end;

function TMyApp.AdvancePos(const StartPos, Amount: Single): Single;
  var pv, v: TG2Vec3;
  var f, l, cp, a, pcp: Single;
  var n0, n1: Integer;
begin
  a := Amount;
  f := Frac(StartPos);
  n0 := Trunc(StartPos) mod Length(Path);
  n1 := (n0 + 1) mod Length(Path);
  v := Path[n1] - Path[n0];
  l := v.Len;
  cp := f * l + a;
  while cp > l do
  begin
    cp := cp - l;
    n0 := n1;
    n1 := (n0 + 1) mod Length(Path);
    v := Path[n1] - Path[n0];
    l := v.Len;
  end;
  Result := n0 + cp / l;
  pv := GetPathPos(StartPos);
  v := GetPathPos(Result);
  l := (v - pv).Len;
  a := a * (Amount / l);
  n0 := Trunc(StartPos) mod Length(Path);
  n1 := (n0 + 1) mod Length(Path);
  v := Path[n1] - Path[n0];
  l := v.Len;
  cp := f * l + a;
  while cp > l do
  begin
    cp := cp - l;
    n0 := n1;
    n1 := (n0 + 1) mod Length(Path);
    v := Path[n1] - Path[n0];
    l := v.Len;
  end;
  Result := n0 + cp / l;
end;

function TMyApp.GetPathPos(const Pos: Single): TG2Vec3;
  var n0, n1, n2, n3, l: Integer;
  var f: Single;
begin
  l := Length(Path);
  f := Frac(Pos);
  n1 := Trunc(Pos); n1 := n1 mod l;
  n0 := n1 - 1; if n0 < 0 then n0 := n0 + l;
  n2 := n1 + 1; n2 := n2 mod l;
  n3 := n1 + 2; n3 := n3 mod l;
  Result := G2Vec3CatmullRom(Path[n0], Path[n1], Path[n2], Path[n3], f);
  //Result := G2LerpVec3(Path[n1], Path[n2], f);
end;
//TMyApp END

end.
